/*
 * master2 - simulate the TREE communication structure with 
 * different slave  
 *
 *
 *
 */
#include <stdio.h>
#include <pvm3.h>
#include <time.h>
#include <unistd.h>
#include <stdlib.h>

#include "common.h"

struct slave_info{
  int tid;
  int runs_left;
  int ready;
};

struct slave_info *info;

int main(int argc,char *argv[])
{
  int myTID;
  int run_todo;
  int times_repeat;
  int noslaves;
  int *tids;
  int i,id,neib;
  
  int total_progress=0;
  int report;
  
  int bytes;
  int type;
  int source;
  int bufid;
  
  time_t astart;
  
  volatile int alldone=0;

  unsigned long finish;
    
  if(argc<4){
    printf("usage: %s <noslaves> <run_todo> <times_repeat>\n",argv[0]);
    exit(1);
  }

  if(sscanf(argv[1],"%i",&noslaves)<=0){
    printf("invalid format of <noslaves>: %s\n",argv[1]);
    exit(1);  
  }
  
  if(sscanf(argv[2],"%i",&run_todo)<=0){
    printf("invalid format of <run_todo>: %s\n",argv[2]);
    exit(1);  
  }
  
  if(sscanf(argv[3],"%i",&times_repeat)<=0){
    printf("invalid format of <times_repeat>: %s\n",argv[3]);
    exit(1);  
  }

  info=(struct slave_info*)malloc(sizeof(struct slave_info)*noslaves);

  tids=(int*)malloc(sizeof(int)*noslaves);

  myTID = pvm_mytid();
  
  if(pvm_spawn("slave2",NULL,PvmTaskDefault,"",noslaves,tids)<noslaves){
    printf("unable to spawn all %i slaves\n",noslaves);
    pvm_exit();
    exit(1);
  };

  for(i=0;i<noslaves;i++){
    info[i].runs_left=times_repeat;
    info[i].tid=tids[i];
    info[i].ready=-1;
  }
  
  finish=run_todo*times_repeat;
  astart=time(NULL);

  printf("master2 TID : %d\n",myTID);
  printf("# of slaves : %i\n",noslaves);
  printf("run_todo    : %i\n",run_todo);
  printf("times_repeat: %i\n",times_repeat);
  printf("finish after: %lu\n",finish);
  printf("astart      : %lu\n",astart);
  
  while(!alldone){
    bufid=pvm_recv(-1, -1);
    pvm_bufinfo(bufid,&bytes,&type,&source);

    /* find the matching entry */
    id=0;
    while(id<noslaves){
      if(source==info[id].tid)break;
      id++;
    }
    if(id<noslaves){
      switch(type){
	case MSG_READY: if(info[id].ready<0){
			  if(id%2){
			    info[id].ready=1;
			    if(times_repeat>0){
			      pvm_initsend(PvmDataDefault);
		              pvm_pkint(&run_todo,1,1);
		              pvm_send(info[id].tid,MSG_JOB);
			      times_repeat--;
			    }
			  }else info[id].ready=0;
			}else{
			  if(id%2)neib=id-1;
			    else neib=id+1;

			  /* a special case here ... */
			  if((id+2==noslaves)&&(noslaves%2==1)){
			    if(times_repeat>0){
			      pvm_initsend(PvmDataDefault);
		              pvm_pkint(&run_todo,1,1);
		              pvm_send(info[id+1].tid,MSG_JOB);
			      times_repeat--;
			    }
			  }
			  if(times_repeat>0){
			    pvm_initsend(PvmDataDefault);
		            pvm_pkint(&run_todo,1,1);
		            pvm_send(info[neib].tid,MSG_JOB);
			    times_repeat--;
			  }
			}
	
			break;
	case MSG_REPORT: pvm_upkint(&report,1,1);
			 total_progress+=report;
			 printf("%lu %u\n",time(NULL)-astart,total_progress);
			 if(total_progress>=finish)alldone=1;
			 break;
	default: printf("got an unknown message id: %i\n",type);
		 break;
      }
    }
  }

  printf("total running time: %lu\n",time(NULL)-astart);

  /* send quit command to all slaves */
  for(i=0;i<noslaves;i++){
    pvm_initsend(PvmDataDefault);
    pvm_send(info[i].tid,MSG_QUIT);
  }  
  pvm_exit();

  return 0;
}
